from dataclasses import dataclass, field
from datetime import datetime
from typing import Literal, Optional

from dacite import from_dict
from httpx import Response

from .Bucket import Bucket
from .client import client
from .exceptions import (
    AuthenticationError,
    ObjectNotFound,
    MatchFailed,
    InternalServerError,
    PermissionError,
)
from .Settings import settings
from .Source import Source
from .Task import Task
from .types import ApiClientProtocol, Schedule


@dataclass
class Project:
    """
    This class implements the structure of Projects of the VC Publisher API.

    :attribute _id: project id
    :atype _id: str
    :attribute name: project name
    :atype name: str
    :attribute bbox: project bounding box
    :atype bbox: list
    :attribute defaultDataBucketId: default data bucket id
    :atype defaultDataBucketId: str
    :attribute createdAt: project creation date
    :atype createdAt: datetime
    :attribute updatedAt: project update date
    :atype updatedAt: datetime
    :attribute createdBy: project creator
    :atype createdBy: User
    :attribute updatedBy: project last updater
    :atype updatedBy: User
    :attribute description: project description
    :atype description: str
    :attribute properties: project properties
    :atype properties: dict
    """

    # Internal attributes
    _api: ApiClientProtocol = field(default=client, init=False, repr=False)
    _endpoint: str = field(init=False, repr=False)

    # required api attributes
    _id: str
    name: str
    bbox: list[float]
    defaultDataBucketId: str  # TODO: converted automatically to Bucket
    createdAt: datetime
    updatedAt: datetime
    createdBy: str | None
    updatedBy: str

    # optional api attributes
    description: Optional[str]
    properties: Optional[dict]

    ############## Data-Buckets ##############
    def create_bucket(
        self,
        name: str,
        description: Optional[str] = None,
        properties: Optional[dict] = None,
    ) -> Bucket:
        """
        Create a bucket for this project.

        :param name: bucket name
        :type name: str
        :param description: optional bucket description
        :type description: str
        :param properties: optional bucket properties
        :type properties: dict
        :return: new bucket
        :rtype: Bucket
        """
        # prepare post request data
        data: dict = {'name': name}
        if description:
            data['description'] = description
        if properties:
            data['properties'] = properties

        # send post request
        response = self._api.post(
            endpoint=self._endpoint + 'data-bucket/', json=data
        )
        # validate response
        match response.status_code:
            case 201:  # Created
                return from_dict(
                    data_class=Bucket,
                    data=response.json(),
                    config=settings.dacite_config,
                )
            case 400:  # Match failed
                raise MatchFailed(
                    f'{response.status_code} - {response.json()["reason"]}'
                )
            case 401:  # Authentication failed
                raise AuthenticationError(
                    f'{response.status_code} - {response.json()["reason"]}'
                )
            case 403:  # Permission denied
                raise PermissionError(
                    f'{response.status_code} - {response.json()["reason"]}'
                )
            case 500:  # Internal server error
                raise InternalServerError(
                    f'{response.status_code} - {response.json()["reason"]}'
                )
            case _:  # All other cases
                raise Exception(
                    f'Failed to create bucket. Response: {response.__dict__}'
                )

    def get_bucket(self, id: str) -> Bucket:
        """
        Get a bucket for this project.

        :param id: bucket id
        :type id: str
        :return: bucket
        :rtype: Bucket
        """
        response = self._api.get(endpoint=self._endpoint + f'data-bucket/{id}/')
        match response.status_code:
            case 200:  # OK
                return from_dict(
                    data_class=Bucket,
                    data=response.json(),
                    config=settings.dacite_config,
                )
            case 400:  # Match failed
                raise MatchFailed(
                    f'{response.status_code} - {response.json()["reason"]}'
                )
            case 401:  # Authentication failed
                raise AuthenticationError(
                    f'{response.status_code} - {response.json()["reason"]}'
                )
            case 403:  # Permission denied
                raise PermissionError(
                    f'{response.status_code} - {response.json()["reason"]}'
                )
            case 404:  # Not found
                raise ObjectNotFound(
                    f'{response.status_code} - {response.json()["reason"]}'
                )
            case _:  # All other cases
                raise Exception(
                    f'Failed to get bucket. Response: {response.__dict__}'
                )

    def update_bucket(
        self,
        id: str,
        name: Optional[str] = None,
        description: Optional[str] = None,
        properties: Optional[dict] = None,
    ) -> Bucket:
        """]
        Update a bucket for this project.

        :param id: bucket id
        :type id: str
        :param name: updated Bucket name
        :type name: Optional[str]
        :param description: updated Bucket description
        :type description: Optional[str]
        :param properties: updated Bucket properties
        :type properties: Optional[dict]
        return: updated bucket
        :rtype: Bucket
        """
        # prepare put request data
        data = {}
        if name:
            data['name'] = name
        if description:
            data['description'] = description
        if properties:
            data['properties'] = properties

        # send put request
        response = self._api.put(
            endpoint=self._endpoint + f'data-bucket/{id}/', json=data
        )
        # validate response
        match response.status_code:
            case 200:  # OK
                return from_dict(
                    data_class=Bucket,
                    data=response.json(),
                    config=settings.dacite_config,
                )
            case 400:  # Match failed
                raise MatchFailed(
                    f'{response.status_code} - {response.json()["reason"]}'
                )
            case 401:  # Authentication failed
                raise AuthenticationError(
                    f'{response.status_code} - {response.json()["reason"]}'
                )
            case 403:  # Permission denied
                raise PermissionError(
                    f'{response.status_code} - {response.json()["reason"]}'
                )
            case 500:  # Internal server error
                raise InternalServerError(
                    f'{response.status_code} - {response.json()["reason"]}'
                )
            case _:  # All other cases
                raise Exception(
                    f'Failed to update bucket. Response: {response.__dict__}'
                )

    def get_buckets(self) -> list[Bucket]:
        """
        Get all buckets for this project.

        :return: list of buckets
        :rtype: list
        """
        response = self._api.get(endpoint=self._endpoint + 'data-buckets/')
        match response.status_code:
            case 200:  # OK
                content = response.json()
                return [
                    from_dict(
                        data_class=Bucket,
                        data=bucket,
                        config=settings.dacite_config,
                    )
                    for bucket in content['items']
                ]
            case 401:  # Authentication failed
                raise AuthenticationError(
                    f'{response.status_code} - {response.json()["reason"]}'
                )
            case 403:  # Permission denied
                raise PermissionError(
                    f'{response.status_code} - {response.json()["reason"]}'
                )
            case 404:  # Not found
                raise ObjectNotFound(
                    f'{response.status_code} - {response.json()["reason"]}'
                )
            case _:  # All other cases
                raise Exception(
                    f'Failed to get buckets. Response: {response.__dict__}'
                )

    ############## Datasources ##############
    def create_source(
        self,
        name: str,
        sourceProperties: dict,  # Union[InternalSource, ExternalSource],
        type: Literal[
            'tileset',
            'tilesetupdate',
            'geojson',
            'oblique',
            'qmesh',
            'meshinmesh',
            'wms',
            'wmts',
            'tms',
            'vectortiles',
            'generic',
        ],
        typeProperties: dict = {},
        description: Optional[str] = None,
        bbox: Optional[list] = None,
        properties: Optional[dict] = None,
    ):
        """
        Create a datasource for this project.

        :param name: datasource name
        :type name: str
        :param sourceProperties: bucket info
        :type sourceProperties: InternalSource | ExternalSource
        :param type: datasource type (e.g. 'tileset', 'geojson', 'wms', etc.)
        :type type: str
        :param typeProperties: datasource type properties
        :type typeProperties: dict
        :param description: optional datasource description
        :type description: str
        :param bbox: optional bounding box
        :type bbox: list
        :param properties: optional datasource properties
        :type properties: dict
        :return: new datasource
        :rtype: Source
        """
        # prepare post request data
        data = {
            'name': name,
            'sourceProperties': sourceProperties,
            'type': type,
            'typeProperties': typeProperties,
        }
        # add optional parameters
        if description:
            data['description'] = description
        if bbox:
            data['bbox'] = bbox
        if properties:
            data['properties'] = properties

        # send post request
        response: Response = self._api.post(
            endpoint=self._endpoint + 'datasource/', json=data
        )
        # validate response
        match response.status_code:
            case 201:
                return from_dict(
                    data_class=Source,
                    data=response.json(),
                    config=settings.dacite_config,
                )
            case 401:
                raise AuthenticationError(
                    f'{response.status_code} - {response.json()["reason"]}'
                )
            case 403:
                raise PermissionError(
                    f'{response.status_code} - {response.json()["reason"]}'
                )
            case _:
                raise Exception(
                    f'Failed to create datasource. Response: {response.__dict__}'
                )

    def get_source(self, id: str):
        """
        Get a datasource for this project.
        """
        response: Response = self._api.get(
            endpoint=self._endpoint + f'datasource/{id}/'
        )
        match response.status_code:
            case 200:
                return from_dict(
                    data_class=Source,
                    data=response.json(),
                    config=settings.dacite_config,
                )
            case 401:
                raise AuthenticationError(
                    f'{response.status_code} - {response.json()["reason"]}'
                )
            case 403:  # Permission denied
                raise PermissionError(
                    f'{response.status_code} - {response.json()["reason"]}'
                )
            case 404:
                raise ObjectNotFound(
                    f'{response.status_code} - {response.json()["reason"]}'
                )
            case _:
                raise Exception(
                    f'Failed to get datasource. Response: {response.__dict__}'
                )

    def get_sources(self):
        """
        Get all datasources for this project.
        """
        response: Response = self._api.get(
            endpoint=self._endpoint + 'datasources/'
        )
        datasources = [
            from_dict(
                data_class=Source,
                data=datasource,
                config=settings.dacite_config,
            )
            for datasource in response.json()['items']
        ]
        return datasources

    ############## Tasks ##############
    def create_task(
        self,
        name: str,
        parameters: dict,
        jobType: str,
        schedule: dict,
        labels: Optional[list[str]] = None,
        tags: Optional[dict] = None,
        debugLevel: Optional[int] = None,  # 0-2
        priority: Optional[float] = None,
        description: Optional[str] = None,
        properties: Optional[dict] = None,
        jobVersion: Optional[str] = None,
    ):
        """
        Create a task for this project.
        """
        # prepare post request data
        data = {
            'name': name,
            'parameters': parameters,
            'jobType': jobType,
            'schedule': schedule,
        }
        # add optional parameters
        if labels:
            data['labels'] = labels
        if tags:
            data['tags'] = tags
        if debugLevel:
            data['debugLevel'] = debugLevel
        if priority:
            data['priority'] = priority
        if description:
            data['description'] = description
        if properties:
            data['properties'] = properties
        if jobVersion:
            data['jobVersion'] = jobVersion

        # send post request
        response = self._api.post(endpoint=self._endpoint + 'task/', json=data)
        # validate response
        match response.status_code:
            case 200:  # OK
                return from_dict(
                    data_class=Task,
                    data=response.json(),
                    config=settings.dacite_config,
                )
            case 201:  # Created
                return from_dict(
                    data_class=Task,
                    data=response.json(),
                    config=settings.dacite_config,
                )
            case 400:  # Match failed
                raise MatchFailed(
                    f'{response.status_code} - {response.json()["reason"]}'
                )
            case 401:  # Authentication failed
                raise AuthenticationError(
                    f'{response.status_code} - {response.json()["reason"]}'
                )
            case 403:
                raise PermissionError(
                    f'{response.status_code} - {response.json()["reason"]}'
                )
            case 500:  # Internal server error
                raise InternalServerError(
                    f'{response.status_code} - {response.json()["reason"]}'
                )
            case _:
                raise Exception(
                    f'Failed to create task. Response: {response.__dict__}'
                )

    def get_task(self, id: str):
        """
        Get a task for this project.
        """
        response: Response = self._api.get(
            endpoint=self._endpoint + f'task/{id}/'
        )
        match response.status_code:
            case 200:
                return from_dict(
                    data_class=Task,
                    data=response.json(),
                    config=settings.dacite_config,
                )
            case 401:
                raise AuthenticationError(
                    f'{response.status_code} - {response.json()["reason"]}'
                )
            case 403:  # Permission denied
                raise PermissionError(
                    f'{response.status_code} - {response.json()["reason"]}'
                )
            case 404:
                raise ObjectNotFound(
                    f'{response.status_code} - {response.json()["reason"]}'
                )
            case _:
                raise Exception(
                    f'Failed to get task. Response: {response.__dict__}'
                )

    def update_task(
        self,
        id: str,
        labels: Optional[list],
        tags: Optional[dict],
        debugLevel: Optional[int],  # 0-2
        priority: Optional[int],
        name: Optional[str],
        description: Optional[str],
        parameters: Optional[dict],
        properties: Optional[dict],
        schedule: Optional[dict],
    ):
        """
        Update a task of this project.

        :param id: task id
        :type id: str
        :param labels: updated task labels
        :type labels: Optional[list]
        :param tags: updated task tags
        :type tags: Optional[dict]
        :param debugLevel: updated task debug level
        :type debugLevel: Optional[int]
        :param priority: updated task priority
        :type priority: Optional[int]
        :param name: updated task name
        :type name: Optional[str]
        :param description: updated task description
        :type description: Optional[str]
        :param parameters: updated task parameters
        :type parameters: Optional[dict]
        :param properties: updated task properties
        :type properties: Optional[dict]
        :param schedule: updated task schedule
        :type schedule: Optional[Schedule]
        :return: updated task
        :rtype: Task
        """
        # prepare put request data
        data = {}
        if labels:
            data['labels'] = labels
        if tags:
            data['tags'] = tags
        if debugLevel:
            data['debugLevel'] = debugLevel
        if priority:
            data['priority'] = priority
        if name:
            data['name'] = name
        if description:
            data['description'] = description
        if parameters:
            data['parameters'] = parameters
        if properties:
            data['properties'] = properties
        if schedule:
            data['schedule'] = schedule

        # send put request
        response: Response = self._api.put(
            endpoint=self._endpoint + f'task/{id}/', json=data
        )
        # validate response
        match response.status_code:
            case 200:  # OK
                return from_dict(
                    data_class=Task,
                    data=response.json(),
                    config=settings.dacite_config,
                )
            case 400:  # Match failed
                raise MatchFailed(
                    f'{response.status_code} - {response.json()["reason"]}'
                )
            case 401:  # Authentication failed
                raise AuthenticationError(
                    f'{response.status_code} - {response.json()["reason"]}'
                )
            case 403:  # Permission denied
                raise PermissionError(
                    f'{response.status_code} - {response.json()["reason"]}'
                )
            case 500:  # Internal server error
                raise InternalServerError(
                    f'{response.status_code} - {response.json()["reason"]}'
                )
            case _:  # All other cases
                raise Exception(
                    f'Failed to update task. Response: {response.__dict__}'
                )

    def get_tasks(self):
        """
        Get all tasks for this project.
        """
        response: Response = self._api.get(endpoint=self._endpoint + 'tasks/')
        tasks = [
            from_dict(data_class=Task, data=task, config=settings.dacite_config)
            for task in response.json()['items']
        ]
        return tasks

    ############## Dunder Methods ##############
    def __post_init__(self):
        """
        Initialize the API endpoint, after the object is created.
        """
        self._endpoint = f'project/{self._id}/'

    def __str__(self):
        """
        String representation of the object as its id.
        """
        return self._id



from httpx import Auth, Request


class BearerAuth(Auth):
    """
    Extends the httpx library with the option to authenticate with a
    bearer token at a Rest API.
    """

    token: str  # Bearer token
    scheme: str = 'Bearer'

    def __init__(self, token: str):
        self.token: str = token

    def auth_flow(self, request: Request):
        # Add the bearer token to the request header
        request.headers['authorization'] = f'{self.scheme} {self.token}'
        yield request


class UserPassAuth(Auth):
    """
    Extends the httpx library with the option to authenticate with a
    user-password tuple at a Rest API.
    """

    user: str  # User
    password: str  # Password

    def __init__(self, user: str, password: str):
        self.user: str = user
        self.password: str = password

    def auth_flow(self, request):
        # Add the user and password to the request header
        request.headers['username'] = self.user
        request.headers['password'] = self.password
        yield request



from dataclasses import dataclass, field
from datetime import datetime
from typing import TYPE_CHECKING, Optional

from .client import client
from .types import ApiClientProtocol

if TYPE_CHECKING:
    from httpx import Response


@dataclass
class Bucket:
    """
    This class implements the structure of Buckets of the VC Publisher API.

    :attribute _id: bucket id
    :atype _id: str
    :attribute name: bucket name
    :atype name: str
    :attribute description: bucket description
    :atype description: str
    :attribute properties: bucket properties
    :atype properties: dict
    :attribute projectId: project id
    :atype projectId: str
    """

    # Internal attributes
    _api: ApiClientProtocol = field(default=client, init=False, repr=False)
    _endpoint: str = field(init=False, repr=False)

    # Required attributes
    _id: str
    createdAt: datetime
    updatedAt: datetime
    createdBy: str
    updatedBy: str
    name: str
    projectId: str

    # Optional attributes
    description: Optional[str] = None
    properties: Optional[dict] = None

    def upload(self, key: str, path: str) -> Response:
        """
        Upload a file to this bucket.

        :param key: key of the file
        :type key: str
        :param path: path of the file
        :type path: str
        :return: Response
        :rtype: Response
        """
        with open(path, 'rb') as file:
            response = self._api.post(
                endpoint=self._endpoint + 'upload/',
                files={key: file},
            )
        return response

    def download(self, key: str):
        """
        Downloads a bucket object or folder as `.tar.gz`.

        This method does not return the file content directly. Instead, it
        returns a generator, which yields the response content in chunks. You
        can iterate over the generator like over httpx streamed responses.
        Example:
            ```
            response = bucket.download()
            with open('download.tar.gz', 'wb') as file:
                for chunk in response.iter_bytes():
                    file.write(chunk)
            ```
        :return:
        """
        return self._api.stream(
            endpoint=self._endpoint + 'download/',
            params={'key': f'/{key}'},
        )

    def download_file(self, key: str):
        """
        Download a bucket object.

        This method does not return the file content directly. Instead, it
        returns a generator, which yields the response content in chunks. You
        can iterate over the generator like over httpx streamed responses.
        Example:
            ```
            response = bucket.download_file()
            with open('download', 'wb') as file:
                for chunk in response.iter_bytes():
                    file.write(chunk)
            ```
        """
        return self._api.stream(
            endpoint=self._endpoint + 'download-file/',
            params={'key': f'/{key}'},
        )

    def delete_object(self, key: str):
        """
        Delete a bucket object.

        :param key: key of the object
        :type key: str
        :return: Response
        :rtype: Response
        """
        return self._api.delete(
            endpoint=self._endpoint + 'object/',
            params={'key': f'/{key}'},
        )

    def delete(self):
        """
        Delete the bucket.

        :return: Response
        :rtype: Response
        """
        return self._api.delete(endpoint=self._endpoint)

    ############## Dunder Methods ##############
    def __post_init__(self):
        """
        Initialize the API endpoint, after the object is created.
        """
        self._endpoint = f'project/{self.projectId}/data-bucket/{self._id}/'

    def __str__(self) -> str:
        return self._id



from dataclasses import dataclass
from typing import Any, Literal, Optional, Protocol

from httpx import Response


class ApiClientProtocol(Protocol):
    def get(self, endpoint: str) -> Response:
        """
        Make a GET Request to the VC Publisher API.

        :return: Response as dict
        """
        ...

    def post(
        self, endpoint: str, data: Optional[dict] = None, json=None, files=None
    ) -> Response:
        """
        Make a POST Request to the VC Publisher API.

        :return: Response as dict
        """
        ...

    def delete(
        self,
        endpoint: str,
        headers: Optional[dict] = None,
        params: Optional[dict] = None,
    ) -> Response:
        """
        Make a DELETE Request to the VC Publisher API.

        :return: Response as dict
        """
        ...

    def put(
        self,
        endpoint: str,
        data: Optional[dict] = None,
        json: Optional[dict] = None,
        files=None,
    ) -> Response:
        """
        Make a PUT Request to the VC Publisher API.
        """
        ...

    def stream(self, endpoint: str, params: Optional[dict] = None) -> Any:
        """
        Make a streaming request to the VC Publisher API.

        :param endpoint: API endpoint
        :type endpoint: str
        :param params: query parameters
        :type params: Optional[dict]
        """
        ...


@dataclass
class SourceProperty:
    """
    SourceProperty is a dataclass that represents one structure of the
    sourceProperties attribute of the Source class.
    """

    type: Literal['external', 'internal']


@dataclass
class ExternalSource(SourceProperty):
    """
    ExternalSource is a dataclass that represents one structure of the
    sourceProperties attribute of the Source class.
    """

    type = 'external'
    url: str


@dataclass
class InternalSource(SourceProperty):
    """
    InternalSource is a dataclass that represents one structure of the
    sourceProperties attribute of the Source class.
    It defines the bucket and key of the internal source.

    :attr dataBucketId: bucket id
    :attr dataBucketKey: bucket key
    :attr urlSuffix: Optional[str]
    """

    type = 'internal'
    dataBucketId: str
    dataBucketKey: str
    urlSuffix: Optional[str] = None


@dataclass
class Schedule:
    """
    Schedule is a dataclass that represents the schedule attribute of the Task
    class.
    There are three types of schedules: immediate, scheduled, and cron.
    For immediate schedules, the type attribute is required.
    For scheduled schedules, the type and scheduled attributes are required.
    For cron schedules, the type and cron attributes are required.
    The suspended attribute is optional for scheduled and cron schedules.

    :attr type: Literal['immediate', 'scheduled', 'cron']
    :attr scheduled: Optional[str]
    :attr cron: Optional[str]
    :attr suspended: Optional[bool]
    """

    type: Literal['immediate', 'scheduled', 'cron']
    # for CronJobs and Scheduled Jobs, the scheduled attribute is required
    scheduled: Optional[str] = None  # datetime in isoformat
    # for CronJobs, the cron attribute is required
    cron: Optional[str] = None
    # for Scheduled Jobs and CronJobs, the suspended attribute is optional
    suspended: Optional[bool] = None

    def to_dict(self) -> dict:
        return self.__dict__



from dacite import from_dict
from httpx import Response

from .client import client
from .exceptions import (
    AuthenticationError,
    InternalServerError,
    MatchFailed,
    ObjectNotFound,
    PermissionError,
)
from .Project import Project
from .Settings import settings
from .types import ApiClientProtocol
from .User import User


def get_project(id: str) -> Project:
    """
    Get project by id

    :param project_id: project id
    :type project_id: str
    :return: project
    :rtype: Project
    """
    api: ApiClientProtocol = client
    response: Response = api.get(
        endpoint=f'project/{id}/',
    )
    match response.status_code:
        case 200:
            return from_dict(
                data_class=Project,
                data=response.json(),
                config=settings.dacite_config,
            )
        case 400:
            raise MatchFailed(
                f'{response.status_code} - {response.json()["reason"]}'
            )
        case 401:
            raise AuthenticationError(
                f'{response.status_code} - {response.json()["reason"]}'
            )
        case 403:
            raise PermissionError(
                f'{response.status_code} - {response.json()["reason"]}'
            )
        case 404:
            raise ObjectNotFound(
                f'{response.status_code} - {response.json()["reason"]}'
            )
        case 500:
            raise InternalServerError(
                f'{response.status_code} - {response.json()["reason"]}'
            )
        case _:
            raise Exception(f'Failed to get project. Response: {response}')


def get_user(user_id: str) -> User:
    """
    Get user by id

    :param user_id: user id
    :type user_id: str
    :return: user
    :rtype: User
    """
    api: ApiClientProtocol = client
    response: Response = api.get(
        endpoint=f'user/{user_id}/',
    )
    match response.status_code:
        case 200:
            return from_dict(
                data_class=User,
                data=response.json(),
                config=settings.dacite_config,
            )
        case _:
            raise Exception(f'Failed to get user. Response: {response}')



class MatchFailed(
    Exception
):
    """
    400 - Match Failed Error.
    """

    def __init__(
        self,
        message,
    ):
        super().__init__(
            message
        )


class AuthenticationError(
    Exception
):
    """
    401 - Authentication Error.
    """

    def __init__(
        self,
        message,
    ):
        super().__init__(
            message
        )


class PermissionError(
    Exception
):
    """
    403 - Permission Error.
    """

    def __init__(
        self,
        message,
    ):
        super().__init__(
            message
        )


class ObjectNotFound(
    Exception
):
    """
    404 - Object not found Exception.
    """

    def __init__(
        self,
        message,
    ):
        super().__init__(
            message
        )


class InternalServerError(
    Exception
):
    """
    500 - Internal Server Error.
    """

    def __init__(
        self,
        message,
    ):
        super().__init__(
            message
        )



import json
import tomllib
from datetime import datetime
from pathlib import Path

from dacite import Config

from .helpers import parse_datetime, parse_source_property
from .types import SourceProperty


class Settings:
    """
    This class is a singleton that loads settings from different sources.

    :attr host: The host of the API
    :atype host: str
    :attr user: The user for the API
    :atype user: str
    :attr password: The password for the API
    :atype password: str
    :attr api_version: The version of the API
    :atype api_version: str
    :attr project_id: The project id
    :atype project_id: str
    """

    _instance = None
    dacite_config = Config(
        type_hooks={
            datetime: parse_datetime,
            SourceProperty: parse_source_property,
        },
    )

    def __new__(cls):
        """
        This is a singleton, so we only create one instance of this class.
        """
        if cls._instance is None:
            cls._instance = super(Settings, cls).__new__(cls)
            cls._instance._load_settings()
        return cls._instance

    def _load_settings(self) -> None:
        """
        This function tries to load settings from different sources.
        """
        # 1. Try to load settings from Django settings.
        if self._load_django_settings():
            print('loaded settings from Django settings')
            return

        # 2. Try to load settings from pyblisher.json.
        elif self._load_json_config():
            print('loaded settings from pyblisher.json')
            return

        # 3. Try to load settings from pyproject.toml.
        elif self._load_toml_config():
            print('loaded settings from pyproject.toml')
            return
        # If no settings are found, raise an AttributeError.
        else:
            print('no settings found')
            raise AttributeError(
                'Found no Settings in pyblisher.toml or Django settings.'
            )

    def _load_django_settings(self) -> bool:
        """
        This function tries to load settings from Django settings.
        """
        try:
            from django.conf import settings

            if hasattr(settings, 'PYBLISHER'):
                for key, value in settings.PYBLISHER.items():
                    setattr(self, key.lower(), value)
                return True
        except ImportError:
            return False
        return False

    def _load_toml_config(self) -> bool:
        """
        This function tries to load settings from a pyblisher.toml file.
        """
        try:
            # Look for pyproject.toml in the current project directory
            config_path = Path.cwd() / 'pyproject.toml'
            if config_path.exists():
                with open(
                    config_path, 'rb'
                ) as f:  # TOML must be read in binary mode
                    config = tomllib.load(f)
                    if 'pyblisher' in config:
                        # Look for [pyblisher] section in pyproject.toml
                        for key, value in config['pyblisher'].items():
                            setattr(self, key.lower(), value)
                    return True
        except Exception as e:
            print(f'Warning: Can not load pyblisher settings. {e}')
        return False

    def _load_json_config(self) -> bool:
        """
        This function tries to load settings from a pyblisher.json file.
        """
        try:
            # Look for pyblisher.json in the current project directory
            config_path = Path.cwd() / 'pyblisher.json'
            if config_path.exists():
                with open(
                    config_path, 'r'
                ) as f:  # JSON must be read in text mode
                    config = json.load(f)
                    for key, value in config.items():
                        setattr(self, key.lower(), value)
                    return True
        except Exception as e:
            print(f'Warning: Can not load pyblisher settings. {e}')
        return False

    def __getattr__(self, name: str):
        """
        This function is called when an attribute is not found.
        """
        raise AttributeError(
            f"'{self.__class__.__name__}' has no attribute '{name}'"
        )

    def __repr__(self):
        """
        This function returns its state in the following format:
            ClassName(attr1=value1, attr2=value2, ...)
        every value is represented with its repr() function
        """
        text = f'{self.__class__.__name__}('
        for key, value in self.__dict__.items():
            text += f'{key}={repr(value)}, '
        text = text[:-2] + ')'
        return text


settings = Settings()
